#     Copyright 2018 Netflix, Inc.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
"""
.. module: policyuniverse.expander_minimizer
    :platform: Unix

.. version:: $$VERSION$$
.. moduleauthor::  Patrick Kelley <<patrickbarrettkelley@gmail.com> @patrickbkelley

"""
from __future__ import print_function
from policyuniverse import all_permissions
import json
import fnmatch
import sys
import copy

policy_headers = ["rolepolicies", "grouppolicies", "userpolicies", "policy"]


def expand_minimize_over_policies(policies, activity, **kwargs):
    for header in policy_headers:
        if header in policies:
            output = {header: {}}
            for policy in policies[header]:
                output[header][policy] = activity(
                    policy=policies[header][policy], **kwargs
                )
            return output

    return activity(policy=policies, **kwargs)


def _get_prefixes_for_action(action):
    """
    :param action: iam:cat
    :return: [ "iam:", "iam:c", "iam:ca", "iam:cat" ]
    """
    (technology, permission) = action.split(":")
    retval = ["{}:".format(technology)]
    phrase = ""
    for char in permission:
        newphrase = "{}{}".format(phrase, char)
        retval.append("{}:{}".format(technology, newphrase))
        phrase = newphrase
    return retval


def _expand_wildcard_action(action):
    """
    :param action: 'autoscaling:*'
    :return: A list of all autoscaling permissions matching the wildcard
    """
    if isinstance(action, list):
        expanded_actions = []
        for item in action:
            expanded_actions.extend(_expand_wildcard_action(item))
        return expanded_actions

    else:
        if "*" in action:
            expanded = [
                expanded_action.lower()
                for expanded_action in all_permissions
                if fnmatch.fnmatchcase(expanded_action.lower(), action.lower())
            ]

            # if we get a wildcard for a tech we've never heard of, just return the wildcard
            if not expanded:
                return [action.lower()]

            return expanded
        return [action.lower()]


def _get_desired_actions_from_statement(statement):
    desired_actions = set()
    actions = _expand_wildcard_action(statement["Action"])

    for action in actions:
        if action not in all_permissions:
            raise Exception(
                "Desired action not found in master permission list. {}".format(action)
            )
        desired_actions.add(action)

    return desired_actions


def _get_denied_prefixes_from_desired(desired_actions):
    denied_actions = all_permissions.difference(desired_actions)
    denied_prefixes = set()
    for denied_action in denied_actions:
        for denied_prefix in _get_prefixes_for_action(denied_action):
            denied_prefixes.add(denied_prefix)

    return denied_prefixes


def _check_min_permission_length(permission, minchars=None):
    if minchars and len(permission) < int(minchars) and permission != "":
        print(
            "Skipping prefix {} because length of {}".format(
                permission, len(permission)
            ),
            file=sys.stderr,
        )
        return True
    return False


def minimize_statement_actions(statement, minchars=None):
    minimized_actions = set()

    if statement["Effect"] != "Allow":
        raise Exception("Minification does not currently work on Deny statements.")

    desired_actions = _get_desired_actions_from_statement(statement)
    denied_prefixes = _get_denied_prefixes_from_desired(desired_actions)

    for action in desired_actions:
        if action in denied_prefixes:
            print("Action is a denied prefix. Action: {}".format(action))
            minimized_actions.add(action)
            continue

        found_prefix = False
        prefixes = _get_prefixes_for_action(action)
        for prefix in prefixes:

            permission = prefix.split(":")[1]
            if _check_min_permission_length(permission, minchars=minchars):
                continue

            if prefix not in denied_prefixes:
                if prefix not in desired_actions:
                    prefix = "{}*".format(prefix)
                minimized_actions.add(prefix)
                found_prefix = True
                break

        if not found_prefix:
            print("Could not suitable prefix. Defaulting to {}".format(prefixes[-1]))
            minimized_actions.add(prefixes[-1])

    # sort the actions
    minimized_actions_list = list(minimized_actions)
    minimized_actions_list.sort()

    return minimized_actions_list


def get_actions_from_statement(statement):
    allowed_actions = set()

    if not type(statement.get("Action", [])) == list:
        statement["Action"] = [statement["Action"]]

    for action in statement.get("Action", []):
        allowed_actions = allowed_actions.union(set(_expand_wildcard_action(action)))

    if not type(statement.get("NotAction", [])) == list:
        statement["NotAction"] = [statement["NotAction"]]

    inverted_actions = set()
    for action in statement.get("NotAction", []):
        inverted_actions = inverted_actions.union(set(_expand_wildcard_action(action)))

    if inverted_actions:
        actions = _invert_actions(inverted_actions)
        allowed_actions = allowed_actions.union(actions)

    return allowed_actions


def _invert_actions(actions):
    from policyuniverse import all_permissions

    return all_permissions.difference(actions)


def expand_policy(policy=None, expand_deny=False):
    # Perform a deepcopy to avoid mutating the input
    result = copy.deepcopy(policy)

    if type(result["Statement"]) is dict:
        result["Statement"] = [result["Statement"]]
    for statement in result["Statement"]:
        if statement["Effect"].lower() == "deny" and not expand_deny:
            continue
        actions = get_actions_from_statement(statement)
        if "NotAction" in statement:
            del statement["NotAction"]
        statement["Action"] = sorted(list(actions))

    return result


def minimize_policy(policy=None, minchars=None):

    str_pol = json.dumps(policy, indent=2)
    size = len(str_pol)

    for statement in policy["Statement"]:
        minimized_actions = minimize_statement_actions(statement, minchars=minchars)
        statement["Action"] = minimized_actions

    str_end_pol = json.dumps(policy, indent=2)
    end_size = len(str_end_pol)

    # print str_end_pol
    print("Start size: {}. End size: {}".format(size, end_size), file=sys.stderr)
    return policy



import json
import os
from collections import defaultdict
from policyuniverse import _action_categories


def translate_aws_action_groups(groups):
    """
    Problem - AWS provides the following five groups:
        - Permissions
        - ReadWrite
        - ListOnly
        - ReadOnly
        - Tagging

    The meaning of these groups was not immediately obvious to me.

    Permissions: ability to modify (create/update/remove) permissions.
    ReadWrite: Indicates a data-plane operation.
    ReadOnly: Always used with ReadWrite. Indicates a read-only data-plane operation.
    ListOnly: Always used with [ReadWrite, ReadOnly]. Indicates an action which
        lists resources, which is a subcategory of read-only data-plane operations.
    Tagging: Always used with ReadWrite. Indicates a permission that can mutate tags.

    So an action with ReadWrite, but without ReadOnly, is a mutating data-plane operation.
    An action with Permission never has any other groups.

    This method will take the AWS categories and translate them to one of the following:

    - List
    - Read
    - Tagging
    - ReadWrite
    - Permissions
    """
    if "Permissions" in groups:
        return "Permissions"
    if "ListOnly" in groups:
        return "List"
    if "ReadOnly" in groups:
        return "Read"
    if "Tagging" in groups:
        return "Tagging"
    if "ReadWrite" in groups:
        return "Write"
    return "Unknown"


def build_action_categories_from_service_data(service_data):
    action_categories = dict()
    for service_name in service_data:
        service_body = service_data[service_name]
        prefix = service_body["prefix"]
        service_actions = service_body["actions"]
        for service_action, service_action_body in service_actions.items():
            key = "{}:{}".format(prefix, service_action.lower())
            action_categories[key] = service_action_body["calculated_action_group"]
    return action_categories


def categories_for_actions(actions):
    """
    Given an iterable of actions, return a mapping of action groups.

    actions: {'ec2:authorizesecuritygroupingress', 'iam:putrolepolicy', 'iam:listroles'}

    Returns:
        {
            'ec2': {'Write'},
            'iam': {'Permissions', 'List'})
        }
    """
    groups = defaultdict(set)
    for action in actions:
        service = action.split(":")[0]
        groups[service].add(_action_categories.get(action))
    return groups


def actions_for_category(category):
    """
    Returns set of actions containing each group passed in.

    Param:
        category must be in {'Permissions', 'List', 'Read', 'Tagging', 'Write'}

    Returns:
        set of matching actions
    """
    actions = set()
    for action, action_category in _action_categories.items():
        if action_category == category:
            actions.add(action)
    return actions



#!/usr/bin/env python

# from distutils.core import setup
from setuptools import setup
import os

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))

tests_require = ["pytest", "coveralls"]
dev_require = ["pre-commit", "black"]

setup(
    name="policyuniverse",
    version="1.3.2.20201201",
    description="Parse and Process AWS IAM Policies, Statements, ARNs, and wildcards.",
    long_description=open(os.path.join(ROOT, "README.md")).read(),
    long_description_content_type="text/markdown",
    author="Patrick Kelley",
    author_email="patrickbarrettkelley@gmail.com",
    url="https://github.com/Netflix-Skunkworks/policyuniverse",
    keywords=[
        "iam",
        "arn",
        "action_groups",
        "condition",
        "policy",
        "statement",
        "wildcard",
    ],
    packages=["policyuniverse"],
    package_data={"policyuniverse": ["data.json"]},
    include_package_data=True,
    zip_safe=False,
    extras_require={"tests": tests_require, "dev": dev_require},
)



#     Copyright 2017 Netflix, Inc.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
"""
.. module: policyuniverse.statement
    :platform: Unix

.. version:: $$VERSION$$
.. moduleauthor::  Patrick Kelley <patrickbarrettkelley@gmail.com> @patrickbkelley

"""
from policyuniverse.arn import ARN
from policyuniverse.expander_minimizer import (
    _expand_wildcard_action,
    get_actions_from_statement,
)
from policyuniverse import logger
from policyuniverse.action_categories import categories_for_actions

import re
from collections import namedtuple


PrincipalTuple = namedtuple("Principal", "category value")
ConditionTuple = namedtuple("Condition", "category value")


class Statement(object):
    def __init__(self, statement):
        self.statement = statement
        self.condition_entries = self._condition_entries()
        self.principals = self._principals()
        self.actions = self._actions()

    @property
    def effect(self):
        return self.statement.get("Effect")

    @property
    def actions_expanded(self):
        return set(get_actions_from_statement(self.statement))

    def _actions(self):
        actions = self.statement.get("Action")
        if not actions:
            return set()
        if not isinstance(actions, list):
            actions = [actions]
        return set(actions)

    def action_summary(self):
        return categories_for_actions(self.actions_expanded)

    def uses_not_principal(self):
        return "NotPrincipal" in self.statement

    @property
    def resources(self):
        if "NotResource" in self.statement:
            return set(["*"])

        resources = self.statement.get("Resource")
        if not isinstance(resources, list):
            resources = [resources]
        return set(resources)

    def whos_allowed(self):
        """Returns set containing any entries from principal and condition section.

        Example:

        statement = Statement(dict(
            Effect='Allow',
            Principal='arn:aws:iam::*:role/Hello',
            Action=['ec2:*'],
            Resource='*',
            Condition={
                'StringLike': {
                    'AWS:SourceOwner': '012345678910'
                }}))

        statement.whos_allowed()
        > set([
        >    PrincipalTuple(category='principal', value='arn:aws:iam::*:role/Hello'),
        >    ConditionTuple(category='account', value='012345678910')])
        """
        who = set()
        for principal in self.principals:
            principal = PrincipalTuple(category="principal", value=principal)
            who.add(principal)
        who = who.union(self.condition_entries)
        return who

    def _principals(self):
        """Extracts all principals from IAM statement.

        Should handle these cases:
        "Principal": "value"
        "Principal": ["value"]
        "Principal": { "AWS": "value" }
        "Principal": { "AWS": ["value", "value"] }
        "Principal": { "Service": "value" }
        "Principal": { "Service": ["value", "value"] }

        Return: Set of principals
        """
        principals = set()
        principal = self.statement.get("Principal", None)
        if not principal:
            # It is possible not to define a principal, AWS ignores these statements.
            return principals

        if isinstance(principal, dict):

            if "AWS" in principal:
                self._add_or_extend(principal["AWS"], principals)

            if "Service" in principal:
                self._add_or_extend(principal["Service"], principals)

            if "Federated" in principal:
                self._add_or_extend(principal["Federated"], principals)

        else:
            self._add_or_extend(principal, principals)

        return principals

    def _add_or_extend(self, value, structure):
        if isinstance(value, list):
            structure.update(set(value))
        else:
            structure.add(value)

    def _condition_entries(self):
        """Extracts any ARNs, Account Numbers, UserIDs, Usernames, CIDRs, VPCs, and VPC Endpoints from a condition block.

        Ignores any negated condition operators like StringNotLike.
        Ignores weak condition keys like referer, date, etc.

        Reason: A condition is meant to limit the principal in a statement.  Often, resource policies use a wildcard principal
        and rely exclusively on the Condition block to limit access.

        We would want to alert if the Condition had no limitations (like a non-existent Condition block), or very weak limitations.  Any negation
        would be weak, and largely equivelant to having no condition block whatsoever.

        The alerting code that relies on this data must ensure the condition has at least one of the following:
        - A limiting ARN
        - Account Identifier
        - AWS Organization Principal Org ID
        - User ID
        - Source IP / CIDR
        - VPC
        - VPC Endpoint

        https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_condition-keys.html
        """
        conditions = list()
        condition = self.statement.get("Condition")
        if not condition:
            return conditions

        key_mapping = {
            "aws:sourcearn": "arn",
            "aws:sourceowner": "account",
            "aws:sourceaccount": "account",
            "aws:principalorgid": "org-id",
            "kms:calleraccount": "account",
            "aws:userid": "userid",
            "aws:sourceip": "cidr",
            "aws:sourcevpc": "vpc",
            "aws:sourcevpce": "vpce",
            # a key for SAML Federation trust policy.
            # https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_create_for-idp_saml.html
            # https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_providers_create_saml_assertions.html
            "saml:aud": "saml-endpoint",
        }

        relevant_condition_operators = [
            re.compile(
                "((ForAllValues|ForAnyValue):)?ARN(Equals|Like)(IfExists)?",
                re.IGNORECASE,
            ),
            re.compile(
                "((ForAllValues|ForAnyValue):)?String(Equals|Like)(IgnoreCase)?(IfExists)?",
                re.IGNORECASE,
            ),
            re.compile(
                "((ForAllValues|ForAnyValue):)?IpAddress(IfExists)?", re.IGNORECASE
            ),
        ]

        for condition_operator in condition.keys():
            if any(
                regex.match(condition_operator)
                for regex in relevant_condition_operators
            ):
                for key, value in condition[condition_operator].items():

                    # ForAllValues and ForAnyValue must be paired with a list.
                    # Otherwise, skip over entries.
                    if not isinstance(
                        value, list
                    ) and condition_operator.lower().startswith("for"):
                        continue

                    if key.lower() in key_mapping:
                        if isinstance(value, list):
                            for v in value:
                                conditions.append(
                                    ConditionTuple(
                                        value=v, category=key_mapping[key.lower()]
                                    )
                                )
                        else:
                            conditions.append(
                                ConditionTuple(
                                    value=value, category=key_mapping[key.lower()]
                                )
                            )

        return conditions

    @property
    def condition_arns(self):
        return self._condition_field("arn")

    @property
    def condition_accounts(self):
        return self._condition_field("account")

    @property
    def condition_orgids(self):
        return self._condition_field("org-id")

    @property
    def condition_userids(self):
        return self._condition_field("userid")

    @property
    def condition_cidrs(self):
        return self._condition_field("cidr")

    @property
    def condition_vpcs(self):
        return self._condition_field("vpc")

    @property
    def condition_vpces(self):
        return self._condition_field("vpce")

    def _condition_field(self, field):
        return set(
            [entry.value for entry in self.condition_entries if entry.category == field]
        )

    def is_internet_accessible(self):
        if self.effect != "Allow":
            return False

        if not self.is_condition_internet_accessible():
            return False

        if self.uses_not_principal():
            return True

        for principal in self.principals:
            if self._arn_internet_accessible(principal):
                return True

        return False

    def is_condition_internet_accessible(self):
        condition_entries = self.condition_entries
        if len(condition_entries) == 0:
            return True

        for entry in condition_entries:
            if self._is_condition_entry_internet_accessible(entry):
                return True

        return False

    def _is_condition_entry_internet_accessible(self, entry):
        if entry.category == "arn":
            return self._arn_internet_accessible(entry.value)

        if entry.category == "userid":
            return self._userid_internet_accessible(entry.value)

        if entry.category == "cidr":
            return self._cidr_internet_accessible(entry.value)

        return "*" in entry.value

    def _cidr_internet_accessible(self, cidr):
        """The caller will want to inspect the CIDRs directly.
        This will only look for /0's.
        """
        return cidr.endswith("/0")

    def _userid_internet_accessible(self, userid):
        # Trailing wildcards are okay for userids:
        # AROAIIIIIIIIIIIIIIIII:*
        if userid.index("*") == len(userid) - 1:
            return False
        return True

    def _arn_internet_accessible(self, arn_input):
        if "*" == arn_input:
            return True

        arn = ARN(arn_input)
        if arn.error:
            logger.warning("Auditor could not parse ARN {arn}.".format(arn=arn_input))
            return "*" in arn_input

        if arn.tech == "s3":
            # S3 ARNs typically don't have account numbers.
            return False

        if not arn.account_number and not arn.service:
            logger.warning(
                "Auditor could not parse Account Number from ARN {arn}.".format(
                    arn=arn_input
                )
            )
            return True

        if arn.account_number == "*":
            return True

        return False
